#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>

// Kernel optimized to minimize warp divergence by refactoring conditional logic
// and ensuring uniform control flow within warps

template <typename scalar_t>
__global__ void warp_divergence_optimized_kernel(
    const scalar_t* __restrict__ input,
    const scalar_t* __restrict__ weight,
    const scalar_t* __restrict__ bias,
    scalar_t* __restrict__ output,
    const int batch_size,
    const int in_features,
    const int out_features,
    const float subtract_value,
    const float multiply_value) {

    extern __shared__ char shared_mem[];
    const int TILE = 16;
    scalar_t* in_tile = (scalar_t*)shared_mem;
    scalar_t* wt_tile = (scalar_t*)&in_tile[TILE * TILE];

    int row = blockIdx.x * TILE + threadIdx.x;
    int col = blockIdx.y * TILE + threadIdx.y;

    scalar_t sum = 0;

    for (int t = 0; t < (in_features + TILE - 1) / TILE; ++t) {
        int load_row = t * TILE + threadIdx.y;
        int load_col = t * TILE + threadIdx.x;

        // Load input tile
        in_tile[threadIdx.x * TILE + threadIdx.y] = (row < batch_size && load_row < in_features) 
            ? input[row * in_features + load_row] : 0;

        // Load weight tile
        wt_tile[threadIdx.y * TILE + threadIdx.x] = (col < out_features && load_col < in_features)
            ? weight[col * in_features + load_col] : 0;

        __syncthreads();

        for (int k = 0; k < TILE; ++k) {
            sum += in_tile[threadIdx.x * TILE + k] * wt_tile[k * TILE + threadIdx.y];
        }
        __syncthreads();
    }

    if (row < batch_size && col < out_features) {
        sum += bias[col];
        sum = (sum - subtract_value) * multiply_value;
        output[row * out_features + col] = max(sum, scalar_t(0));
    }
}

torch::Tensor forward(
    torch::Tensor input,
    torch::Tensor weight,
    torch::Tensor bias,
    float subtract_value,
    float multiply_value) {
    
    auto batch_size = input.size(0);
    auto in_features = input.size(1);
    auto out_features = weight.size(0);

    auto output = torch::empty({batch_size, out_features}, input.options());

    const int TILE = 16;
    dim3 threads(TILE, TILE);
    dim3 blocks(
        (batch_size + TILE - 1) / TILE,
        (out_features + TILE - 1) / TILE
    );

    size_t shared_mem_size = 2 * TILE * TILE * sizeof(float);

    AT_DISPATCH_FLOATING_TYPES(input.type(), "warp_divergence_optimized_kernel", ([&] {
        warp_divergence_optimized_kernel<scalar_t><<<blocks, threads, shared_mem_size>>>(
            input.data_ptr<scalar_t>(),
            weight.data_ptr<scalar_t>(),
            bias.data_ptr<scalar_t>(),
            output.data_ptr<scalar_t>(),
            batch_size,
            in_features,
            out_features,
            subtract_value,
            multiply_value
        );
    }));

    return output;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Kernel optimized for warp divergence");
}